# --------------------------------------------------------------------------------------
# Copyright (c) 2021, Nucleic Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# --------------------------------------------------------------------------------------

from typing import Any, Iterable, NoReturn, Tuple, type_check_only

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal  # type: ignore

# --- Exceptions

class BadRequiredStrength(Exception): ...
class DuplicateConstraint(Exception): ...
class DuplicateEditVariable(Exception): ...
class UnknownConstraint(Exception): ...
class UnknownEditVariable(Exception): ...
class UnsatisfiableConstraint(Exception): ...

# Types
@type_check_only
class Strength:
    @property
    def weak(self) -> float: ...
    @property
    def medium(self) -> float: ...
    @property
    def strong(self) -> float: ...
    @property
    def required(self) -> float: ...
    def create(
        self,
        a: int | float,
        b: int | float,
        c: int | float,
        weight: int | float = 1.0,
        /,
    ) -> float: ...

# This is meant as a singleton and users should not access the Strength type.
strength: Strength

class Variable:
    """Variable to express a constraint in a solver."""

    __hash__: None  # type: ignore
    def __init__(self, name: str = "", context: Any = None, /) -> None: ...
    def name(self) -> str:
        """Get the name of the variable."""
        ...
    def setName(self, name: str, /) -> Any:
        """Set the name of the variable."""
        ...
    def value(self) -> float:
        """Get the current value of the variable."""
        ...
    def context(self) -> Any:
        """Get the context object associated with the variable."""
        ...
    def setContext(self, context: Any, /) -> Any:
        """Set the context object associated with the variable."""
        ...
    def __neg__(self) -> Term: ...
    def __add__(self, other: float | Variable | Term | Expression) -> Expression: ...
    def __radd__(self, other: float | Variable | Term | Expression) -> Expression: ...
    def __sub__(self, other: float | Variable | Term | Expression) -> Expression: ...
    def __rsub__(self, other: float | Variable | Term | Expression) -> Expression: ...
    def __mul__(self, other: float) -> Term: ...
    def __rmul__(self, other: float) -> Term: ...
    def __truediv__(self, other: float) -> Term: ...
    def __rtruediv__(self, other: float) -> Term: ...
    def __eq__(self, other: float | Variable | Term | Expression) -> Constraint: ...  # type: ignore
    def __ge__(self, other: float | Variable | Term | Expression) -> Constraint: ...
    def __le__(self, other: float | Variable | Term | Expression) -> Constraint: ...
    def __ne__(self, other: Any) -> NoReturn: ...
    def __gt__(self, other: Any) -> NoReturn: ...
    def __lt__(self, other: Any) -> NoReturn: ...

class Term:
    """Product of a variable by a constant pre-factor."""

    __hash__: None  # type: ignore
    def __init__(
        self, variable: Variable, coefficient: int | float = 1.0, /  # noqa
    ) -> None: ...
    def coefficient(self) -> float:
        """Get the coefficient for the term."""
        ...
    def variable(self) -> Variable:
        """Get the variable for the term."""
        ...
    def value(self) -> float:
        """Get the value for the term."""
        ...
    def __neg__(self) -> Term: ...
    def __add__(self, other: float | Variable | Term | Expression) -> Expression: ...
    def __radd__(self, other: float | Variable | Term | Expression) -> Expression: ...
    def __sub__(self, other: float | Variable | Term | Expression) -> Expression: ...
    def __rsub__(self, other: float | Variable | Term | Expression) -> Expression: ...
    def __mul__(self, other: float) -> Term: ...
    def __rmul__(self, other: float) -> Term: ...
    def __truediv__(self, other: float) -> Term: ...
    def __rtruediv__(self, other: float) -> Term: ...
    def __eq__(self, other: float | Variable | Term | Expression) -> Constraint: ...  # type: ignore
    def __ge__(self, other: float | Variable | Term | Expression) -> Constraint: ...
    def __le__(self, other: float | Variable | Term | Expression) -> Constraint: ...
    def __ne__(self, other: Any) -> NoReturn: ...
    def __gt__(self, other: Any) -> NoReturn: ...
    def __lt__(self, other: Any) -> NoReturn: ...

class Expression:
    """Sum of terms and an additional constant."""

    __hash__: None  # type: ignore
    def __init__(
        self, terms: Iterable[Term], constant: int | float = 0.0, /  # noqa
    ) -> None: ...
    def constant(self) -> float:
        "" "Get the constant for the expression." ""
        ...
    def terms(self) -> Tuple[Term, ...]:
        """Get the tuple of terms for the expression."""
        ...
    def value(self) -> float:
        """Get the value for the expression."""
        ...
    def __neg__(self) -> Expression: ...
    def __add__(self, other: float | Variable | Term | Expression) -> Expression: ...
    def __radd__(self, other: float | Variable | Term | Expression) -> Expression: ...
    def __sub__(self, other: float | Variable | Term | Expression) -> Expression: ...
    def __rsub__(self, other: float | Variable | Term | Expression) -> Expression: ...
    def __mul__(self, other: float) -> Expression: ...
    def __rmul__(self, other: float) -> Expression: ...
    def __truediv__(self, other: float) -> Expression: ...
    def __rtruediv__(self, other: float) -> Expression: ...
    def __eq__(self, other: float | Variable | Term | Expression) -> Constraint: ...  # type: ignore
    def __ge__(self, other: float | Variable | Term | Expression) -> Constraint: ...
    def __le__(self, other: float | Variable | Term | Expression) -> Constraint: ...
    def __ne__(self, other: Any) -> NoReturn: ...
    def __gt__(self, other: Any) -> NoReturn: ...
    def __lt__(self, other: Any) -> NoReturn: ...

class Constraint:
    def __init__(
        self,
        expression: Expression,
        op: Literal["=="] | Literal["<="] | Literal[">="],
        strength: float
        | Literal["weak"]
        | Literal["medium"]
        | Literal["strong"]
        | Literal["required"] = "required",
        /,
    ) -> None: ...
    def expression(self) -> Expression:
        """Get the expression object for the constraint."""
        ...
    def op(self) -> Literal["=="] | Literal["<="] | Literal[">="]:
        """Get the relational operator for the constraint."""
        ...
    def strength(self) -> float:
        """Get the strength for the constraint."""
        ...
    def __or__(
        self,
        other: float
        | Literal["weak"]
        | Literal["medium"]
        | Literal["strong"]
        | Literal["required"],
    ) -> Constraint: ...
    def __ror__(
        self,
        other: float
        | Literal["weak"]
        | Literal["medium"]
        | Literal["strong"]
        | Literal["required"],
    ) -> Constraint: ...

class Solver:
    """Kiwi solver class."""

    def __init__(self) -> None: ...
    def addConstraint(self, constraint: Constraint, /) -> None:
        """Add a constraint to the solver."""
        ...
    def removeConstraint(self, constraint: Constraint, /) -> None:
        """Remove a constraint from the solver."""
        ...
    def hasConstraint(self, constraint: Constraint, /) -> bool:
        """Check whether the solver contains a constraint."""
        ...
    def addEditVariable(
        self,
        variable: Variable,
        strength: float
        | Literal["weak"]
        | Literal["medium"]
        | Literal["strong"]
        | Literal["required"],
        /,
    ) -> None:
        """Add an edit variable to the solver."""
        ...
    def removeEditVariable(self, variable: Variable, /) -> None:
        """Remove an edit variable from the solver."""
        ...
    def hasEditVariable(self, variable: Variable, /) -> bool:
        """Check whether the solver contains an edit variable."""
        ...
    def suggestValue(self, variable: Variable, value: int | float, /) -> None:
        """Suggest a desired value for an edit variable."""
        ...
    def updateVariables(self) -> None:
        """Update the values of the solver variables."""
        ...
    def reset(self) -> None:
        """Reset the solver to the initial empty starting condition."""
        ...
    def dump(self) -> None:
        """Dump a representation of the solver internals to stdout."""
        ...
    def dumps(self) -> str:
        """Dump a representation of the solver internals to a string."""
        ...



# --------------------------------------------------------------------------------------
# Copyright (c) 2014-2021, Nucleic Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# --------------------------------------------------------------------------------------
import gc
import re

import pytest

from kiwisolver import Constraint, Variable, strength


@pytest.mark.parametrize("op", ("==", "<=", ">="))
def test_constraint_creation(op) -> None:
    """Test constraints creation and methods."""
    v = Variable("foo")
    c = Constraint(v + 1, op)

    assert c.strength() == strength.required and c.op() == op
    e = c.expression()
    t = e.terms()
    assert (
        e.constant() == 1
        and len(t) == 1
        and t[0].variable() is v
        and t[0].coefficient() == 1
    )

    constraint_format = r"1 \* foo \+ 1 %s 0 | strength = 1.001e\+[0]+9" % op
    assert re.match(constraint_format, str(c))

    for s in ("weak", "medium", "strong", "required"):
        # Not an exact literal...
        c = Constraint(v + 1, op, s)  # type: ignore
        assert c.strength() == getattr(strength, s)

    # Ensure we test garbage collection.
    del c
    gc.collect()


def test_constraint_creation2() -> None:
    """Test for errors in Constraints creation."""
    v = Variable("foo")

    with pytest.raises(TypeError) as excinfo:
        Constraint(1, "==")  # type: ignore
    assert "Expression" in excinfo.exconly()

    with pytest.raises(TypeError) as excinfo:
        Constraint(v + 1, 1)  # type: ignore
    assert "str" in excinfo.exconly()

    with pytest.raises(ValueError) as excinfo2:
        Constraint(v + 1, "!=")  # type: ignore
    assert "relational operator" in excinfo2.exconly()


@pytest.mark.parametrize("op", ("==", "<=", ">="))
def test_constraint_repr(op) -> None:
    """Test the repr method of a constraint object."""
    v = Variable("foo")
    c = Constraint(v + 1, op)

    assert op in repr(c)


def test_constraint_or_operator() -> None:
    """Test modifying a constraint strength using the | operator."""
    v = Variable("foo")
    c = Constraint(v + 1, "==")

    for s in ("weak", "medium", "strong", "required", strength.create(1, 1, 0)):
        c2 = c | s  # type: ignore
        if isinstance(s, str):
            assert c2.strength() == getattr(strength, s)
        else:
            assert c2.strength() == s

    with pytest.raises(ValueError):
        c | "unknown"  # type: ignore



# --------------------------------------------------------------------------------------
# Copyright (c) 2013-2022, Nucleic Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# --------------------------------------------------------------------------------------
import os

from setuptools import Extension, setup

try:
    from cppy import CppyBuildExt
except ImportError as e:
    raise RuntimeError(
        "Missing setup required dependencies: cppy. "
        "Installing through pip as recommended ensure one never hits this issue."
    ) from e

# Before releasing the version needs to be updated in kiwi/version.h, if the changes
# are not limited to the solver.

# Use the env var KIWI_DISABLE_FH4 to disable linking against VCRUNTIME140_1.dll

if "KIWI_DISABLE_FH4" in os.environ:
    os.environ.setdefault("CPPY_DISABLE_FH4", "1")

ext_modules = [
    Extension(
        "kiwisolver",
        [
            "py/kiwisolver.cpp",
            "py/constraint.cpp",
            "py/expression.cpp",
            "py/solver.cpp",
            "py/strength.cpp",
            "py/term.cpp",
            "py/variable.cpp",
        ],
        include_dirs=["."],
        language="c++",
    ),
]


setup(
    # FIXME remove once setuptool supports PEP 621
    name="kiwisolver",
    author="The Nucleic Development Team",
    author_email="sccolbert@gmail.com",
    url="https://github.com/nucleic/kiwi",
    description="A fast implementation of the Cassowary constraint solver",
    long_description=open("README.rst").read(),
    license="BSD",
    classifiers=[
        # https://pypi.org/pypi?%3Aaction=list_classifiers
        "License :: OSI Approved :: BSD License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: Implementation :: CPython",
        "Programming Language :: Python :: Implementation :: PyPy",
    ],
    python_requires=">=3.7",
    # FIXME end of remove once setuptool supports PEP 621
    setup_requires=[
        "setuptools>=42",
        "wheel",
        "setuptools_scm[toml]>=3.4.3",
        "cppy>=1.2.0",
    ],
    install_requires=["typing_extensions;python_version<'3.8'"],
    ext_modules=ext_modules,
    cmdclass={"build_ext": CppyBuildExt},
    package_data={"py": ["py.typed", "*.pyi"]},
)



# --------------------------------------------------------------------------------------
# Copyright (c) 2014-2021, Nucleic Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# --------------------------------------------------------------------------------------
import gc
import math
import operator
from typing import Tuple

import pytest

from kiwisolver import Constraint, Expression, Term, Variable, strength


def test_term_creation() -> None:
    """Test the Term constructor."""
    v = Variable("foo")
    t = Term(v)
    assert t.variable() is v
    assert t.coefficient() == 1

    t = Term(v, 100)
    assert t.variable() is v
    assert t.coefficient() == 100

    assert str(t) == "100 * foo"

    with pytest.raises(TypeError) as excinfo:
        Term("")  # type: ignore
    assert "Variable" in excinfo.exconly()

    # ensure we test garbage collection
    del t
    gc.collect()


@pytest.fixture()
def terms():
    """Terms for testing."""
    v = Variable("foo")
    v2 = Variable("bar")
    t = Term(v, 10)
    t2 = Term(v2)

    return t, t2, v, v2


def test_term_neg(terms: Tuple[Term, Term, Variable, Variable]) -> None:
    """Test neg on a term."""
    t, _, v, _ = terms

    neg = -t
    assert isinstance(neg, Term)
    assert neg.variable() is v and neg.coefficient() == -10


def test_term_mul(terms: Tuple[Term, Term, Variable, Variable]) -> None:
    """Test term multiplications"""
    t, _, v, _ = terms

    for mul in (t * 2, 2.0 * t):
        assert isinstance(mul, Term)
        assert mul.variable() is v and mul.coefficient() == 20

    with pytest.raises(TypeError):
        t * v  # type: ignore


def test_term_div(terms: Tuple[Term, Term, Variable, Variable]) -> None:
    """Test term divisions."""
    t, _, v, v2 = terms

    div = t / 2
    assert isinstance(div, Term)
    assert div.variable() is v and div.coefficient() == 5

    with pytest.raises(TypeError):
        t / v2  # type: ignore

    with pytest.raises(ZeroDivisionError):
        t / 0


def test_term_add(terms: Tuple[Term, Term, Variable, Variable]) -> None:
    """Test term additions."""
    t, t2, v, v2 = terms

    for add in (t + 2, 2.0 + t):
        assert isinstance(add, Expression)
        assert add.constant() == 2
        terms_ = add.terms()
        assert (
            len(terms_) == 1
            and terms[0].variable() is v
            and terms_[0].coefficient() == 10
        )

    for add2, order in zip((t + v2, v2 + t), ((0, 1), (1, 0))):
        assert isinstance(add2, Expression)
        assert add2.constant() == 0
        terms_ = add2.terms()
        assert (
            len(terms_) == 2
            and terms_[order[0]].variable() is v
            and terms_[order[0]].coefficient() == 10
            and terms_[order[1]].variable() is v2
            and terms_[order[1]].coefficient() == 1
        )

    add2 = t + t2
    assert isinstance(add2, Expression)
    assert add2.constant() == 0
    terms_ = add2.terms()
    assert (
        len(terms_) == 2
        and terms_[0].variable() is v
        and terms_[0].coefficient() == 10
        and terms_[1].variable() is v2
        and terms_[1].coefficient() == 1
    )


def test_term_sub(terms: Tuple[Term, Term, Variable, Variable]) -> None:
    """Test term substractions."""
    t, t2, v, v2 = terms

    for sub, diff in zip((t - 2, 2.0 - t), (-2, 2)):
        assert isinstance(sub, Expression)
        assert sub.constant() == diff
        terms_ = sub.terms()
        assert (
            len(terms_) == 1
            and terms[0].variable() is v
            and terms_[0].coefficient() == -math.copysign(10, diff)
        )

    for sub2, order in zip((t - v2, v2 - t), ((0, 1), (1, 0))):
        assert isinstance(sub2, Expression)
        assert sub2.constant() == 0
        terms_ = sub2.terms()
        assert (
            len(terms_) == 2
            and terms_[order[0]].variable() is v
            and terms_[order[0]].coefficient() == 10 * (-1) ** order[0]
            and terms_[order[1]].variable() is v2
            and terms_[order[1]].coefficient() == -1 * (-1) ** order[0]
        )

    sub2 = t - t2
    assert isinstance(sub2, Expression)
    assert sub2.constant() == 0
    terms_ = sub2.terms()
    assert (
        len(terms_) == 2
        and terms_[0].variable() is v
        and terms_[0].coefficient() == 10
        and terms_[1].variable() is v2
        and terms_[1].coefficient() == -1
    )


@pytest.mark.parametrize(
    "op, symbol",
    [
        (operator.le, "<="),
        (operator.eq, "=="),
        (operator.ge, ">="),
        (operator.lt, None),
        (operator.ne, None),
        (operator.gt, None),
    ],
)
def test_term_rich_compare_operations(op, symbol):
    """Test using comparison on variables."""
    v = Variable("foo")
    v2 = Variable("bar")
    t1 = Term(v, 10)
    t2 = Term(v2, 20)

    if symbol is not None:
        c = op(t1, t2 + 1)
        assert isinstance(c, Constraint)
        e = c.expression()
        t = e.terms()
        assert len(t) == 2
        if t[0].variable() is not v:
            t = (t[1], t[0])
        assert (
            t[0].variable() is v
            and t[0].coefficient() == 10
            and t[1].variable() is v2
            and t[1].coefficient() == -20
        )
        assert e.constant() == -1
        assert c.op() == symbol and c.strength() == strength.required

    else:
        with pytest.raises(TypeError):
            op(t1, t2)
